@@ -14,6 +14,7 @@
 
 import copy
 import datetime
+import jsonschema
 import logging
 import pytz
 import six
@@ -994,6 +995,15 @@ class _TestObject(object):
         bar.foo = 1
         self.assertEqual(set(['bar']), obj.obj_what_changed())
 
+    def test_changed_with_bogus_field(self):
+        obj = MyObj()
+        obj.foo = 123
+        # Add a bogus field name to the changed list, as could be the
+        # case if we're sent some broken primitive from another node.
+        obj._changed_fields.add('does_not_exist')
+        self.assertEqual(set(['foo']), obj.obj_what_changed())
+        self.assertEqual({'foo': 123}, obj.obj_get_changes())
+
     def test_static_result(self):
         obj = MyObj.query(self.context)
         self.assertEqual(obj.bar, 'bar')
@@ -1334,6 +1344,20 @@ class _TestObject(object):
                 primitive['rel_objects'][0]['versioned_object.data'],
                 '1.2', version_manifest=manifest)
 
+    def test_obj_make_compatible_removes_field_cleans_changes(self):
+        @base.VersionedObjectRegistry.register_if(False)
+        class TestObject(base.VersionedObject):
+            VERSION = '1.1'
+            fields = {'foo': fields.StringField(),
+                      'bar': fields.StringField()}
+
+            def obj_make_compatible(self, primitive, target_version):
+                del primitive['bar']
+
+        obj = TestObject(foo='test1', bar='test2')
+        prim = obj.obj_to_primitive('1.0')
+        self.assertEqual(['foo'], prim['versioned_object.changes'])
+
     def test_delattr(self):
         obj = MyObj(bar='foo')
         del obj.bar
@@ -2042,35 +2066,26 @@ class TestObjectSerializer(_BaseTestCase):
 
 
 class TestSchemaGeneration(test.TestCase):
-    class FakeFieldType(fields.FieldType):
-        pass
+    @base.VersionedObjectRegistry.register
+    class FakeObject(base.VersionedObject):
+        fields = {
+            'a_boolean': fields.BooleanField(nullable=True),
+        }
 
-    def setUp(self):
-        super(TestSchemaGeneration, self).setUp()
-
-        self.nonNullableField = fields.Field(self.FakeFieldType)
-        self.nullableField = fields.Field(self.FakeFieldType)
-
-        class TestObject(base.VersionedObject):
-            fields = {'foo': self.nonNullableField,
-                      'bar': self.nullableField}
-
-        self.test_class = TestObject
-
-        self.nonNullableField.get_schema = \
-            mock.Mock(return_value={'type': ['fake']})
-        self.nullableField.get_schema = \
-            mock.Mock(return_value={'type': ['fake', 'null']})
-        self.test_class.obj_name = mock.Mock(return_value='TestObject')
+    @base.VersionedObjectRegistry.register
+    class FakeComplexObject(base.VersionedObject):
+        fields = {
+            'a_dict': fields.DictOfListOfStringsField(),
+            'an_obj': fields.ObjectField('FakeObject', nullable=True),
+            'list_of_objs': fields.ListOfObjectsField('FakeObject'),
+        }
 
     def test_to_json_schema(self):
-        schema = self.test_class.to_json_schema()
-        self.nonNullableField.get_schema.assert_called_once_with()
-        self.nullableField.get_schema.assert_called_once_with()
+        schema = self.FakeObject.to_json_schema()
         self.assertEqual({
             '$schema': 'http://json-schema.org/draft-04/schema#',
-            'title': 'TestObject',
-            'type': 'object',
+            'title': 'FakeObject',
+            'type': ['object'],
             'properties': {
                 'versioned_object.namespace': {
                     'type': 'string'
@@ -2089,18 +2104,113 @@ class TestSchemaGeneration(test.TestCase):
                 },
                 'versioned_object.data': {
                     'type': 'object',
-                    'description': 'fields of TestObject',
+                    'description': 'fields of FakeObject',
                     'properties': {
-                        'foo': {'type': ['fake']},
-                        'bar': {'type': ['fake', 'null']}
+                        'a_boolean': {
+                            'readonly': False,
+                            'type': ['boolean', 'null']},
                     },
-                    'required': ['bar', 'foo'],
                 },
             },
             'required': ['versioned_object.namespace', 'versioned_object.name',
                          'versioned_object.version', 'versioned_object.data']
         }, schema)
 
+        jsonschema.validate(self.FakeObject(a_boolean=True).obj_to_primitive(),
+                            self.FakeObject.to_json_schema())
+
+    def test_to_json_schema_complex_object(self):
+        schema = self.FakeComplexObject.to_json_schema()
+        expected_schema = {
+            '$schema': 'http://json-schema.org/draft-04/schema#',
+            'properties': {
+                'versioned_object.changes':
+                    {'items': {'type': 'string'}, 'type': 'array'},
+                'versioned_object.data': {
+                    'description': 'fields of FakeComplexObject',
+                    'properties': {
+                        'a_dict': {
+                            'readonly': False,
+                            'type': ['object'],
+                            'additionalProperties': {
+                                'type': ['array'],
+                                'readonly': False,
+                                'items': {
+                                    'type': ['string'],
+                                    'readonly': False}}},
+                        'an_obj': {
+                            'properties': {
+                                'versioned_object.changes':
+                                    {'items': {'type': 'string'},
+                                     'type': 'array'},
+                                'versioned_object.data': {
+                                    'description': 'fields of FakeObject',
+                                    'properties':
+                                        {'a_boolean': {'readonly': False,
+                                         'type': ['boolean', 'null']}},
+                                    'type': 'object'},
+                                'versioned_object.name': {'type': 'string'},
+                                'versioned_object.namespace':
+                                    {'type': 'string'},
+                                'versioned_object.version':
+                                    {'type': 'string'}},
+                                'readonly': False,
+                                'required': ['versioned_object.namespace',
+                                             'versioned_object.name',
+                                             'versioned_object.version',
+                                             'versioned_object.data'],
+                                'type': ['object', 'null']},
+                        'list_of_objs': {
+                            'items': {
+                                'properties': {
+                                    'versioned_object.changes':
+                                        {'items': {'type': 'string'},
+                                         'type': 'array'},
+                                    'versioned_object.data': {
+                                        'description': 'fields of FakeObject',
+                                        'properties': {
+                                            'a_boolean': {
+                                                'readonly': False,
+                                                'type': ['boolean', 'null']}},
+                                            'type': 'object'},
+                                    'versioned_object.name':
+                                        {'type': 'string'},
+                                    'versioned_object.namespace':
+                                        {'type': 'string'},
+                                    'versioned_object.version':
+                                        {'type': 'string'}},
+                                'readonly': False,
+                                'required': ['versioned_object.namespace',
+                                             'versioned_object.name',
+                                             'versioned_object.version',
+                                             'versioned_object.data'],
+                                'type': ['object']},
+                            'readonly': False,
+                            'type': ['array']}},
+                    'required': ['a_dict', 'list_of_objs'],
+                    'type': 'object'},
+                'versioned_object.name': {'type': 'string'},
+                'versioned_object.namespace': {'type': 'string'},
+                'versioned_object.version': {'type': 'string'}},
+            'required': ['versioned_object.namespace',
+                         'versioned_object.name',
+                         'versioned_object.version',
+                         'versioned_object.data'],
+            'title': 'FakeComplexObject',
+            'type': ['object']}
+        self.assertEqual(expected_schema, schema)
+
+        fake_obj = self.FakeComplexObject(
+            a_dict={'key1': ['foo', 'bar'],
+                    'key2': ['bar', 'baz']},
+            an_obj=self.FakeObject(a_boolean=True),
+            list_of_objs=[self.FakeObject(a_boolean=False),
+                          self.FakeObject(a_boolean=True),
+                          self.FakeObject(a_boolean=False)])
+
+        primitives = fake_obj.obj_to_primitive()
+        jsonschema.validate(primitives, schema)
+
 
 class TestNamespaceCompatibility(test.TestCase):
     def setUp(self):
@@ -2228,3 +2338,92 @@ class TestUtilityMethods(test.TestCase):
                           'TestChild': '2.34',
                           'TestChildTwo': '4.56'},
                          tree)
+
+
+class TestListObjectConcat(test.TestCase):
+    def test_list_object_concat(self):
+        @base.VersionedObjectRegistry.register_if(False)
+        class MyList(base.ObjectListBase, base.VersionedObject):
+            fields = {'objects': fields.ListOfObjectsField('MyOwnedObject')}
+
+        values = [1, 2, 42]
+
+        list1 = MyList(objects=[MyOwnedObject(baz=values[0]),
+                                MyOwnedObject(baz=values[1])])
+        list2 = MyList(objects=[MyOwnedObject(baz=values[2])])
+
+        concat_list = list1 + list2
+        for idx, obj in enumerate(concat_list):
+            self.assertEqual(values[idx], obj.baz)
+
+        # Assert that the original lists are unmodified
+        self.assertEqual(2, len(list1.objects))
+        self.assertEqual(1, list1.objects[0].baz)
+        self.assertEqual(2, list1.objects[1].baz)
+        self.assertEqual(1, len(list2.objects))
+        self.assertEqual(42, list2.objects[0].baz)
+
+    def test_list_object_concat_fails_different_objects(self):
+        @base.VersionedObjectRegistry.register_if(False)
+        class MyList(base.ObjectListBase, base.VersionedObject):
+            fields = {'objects': fields.ListOfObjectsField('MyOwnedObject')}
+
+        @base.VersionedObjectRegistry.register_if(False)
+        class MyList2(base.ObjectListBase, base.VersionedObject):
+            fields = {'objects': fields.ListOfObjectsField('MyOwnedObject')}
+
+        list1 = MyList(objects=[MyOwnedObject(baz=1)])
+        list2 = MyList2(objects=[MyOwnedObject(baz=2)])
+
+        def add(x, y):
+            return x + y
+
+        self.assertRaises(TypeError, add, list1, list2)
+        # Assert that the original lists are unmodified
+        self.assertEqual(1, len(list1.objects))
+        self.assertEqual(1, len(list2.objects))
+        self.assertEqual(1, list1.objects[0].baz)
+        self.assertEqual(2, list2.objects[0].baz)
+
+    def test_list_object_concat_fails_extra_fields(self):
+        @base.VersionedObjectRegistry.register_if(False)
+        class MyList(base.ObjectListBase, base.VersionedObject):
+            fields = {'objects': fields.ListOfObjectsField('MyOwnedObject'),
+                      'foo': fields.IntegerField(nullable=True)}
+
+        list1 = MyList(objects=[MyOwnedObject(baz=1)])
+        list2 = MyList(objects=[MyOwnedObject(baz=2)])
+
+        def add(x, y):
+            return x + y
+
+        self.assertRaises(TypeError, add, list1, list2)
+        # Assert that the original lists are unmodified
+        self.assertEqual(1, len(list1.objects))
+        self.assertEqual(1, len(list2.objects))
+        self.assertEqual(1, list1.objects[0].baz)
+        self.assertEqual(2, list2.objects[0].baz)
+
+    def test_builtin_list_add_fails(self):
+        @base.VersionedObjectRegistry.register_if(False)
+        class MyList(base.ObjectListBase, base.VersionedObject):
+            fields = {'objects': fields.ListOfObjectsField('MyOwnedObject')}
+
+        list1 = MyList(objects=[MyOwnedObject(baz=1)])
+
+        def add(obj):
+            return obj + []
+
+        self.assertRaises(TypeError, add, list1)
+
+    def test_builtin_list_radd_fails(self):
+        @base.VersionedObjectRegistry.register_if(False)
+        class MyList(base.ObjectListBase, base.VersionedObject):
+            fields = {'objects': fields.ListOfObjectsField('MyOwnedObject')}
+
+        list1 = MyList(objects=[MyOwnedObject(baz=1)])
+
+        def add(obj):
+            return [] + obj
+
+        self.assertRaises(TypeError, add, list1)
